﻿
function init() {

    loadPerfDiag();
    genDefaultHighResolutionDiskUsageGraph();
    genHighProcessorsUsagePeriods();
    genHighMemoryUsagePeriods();
    genHighCpuConsumersGraphs();
    genHighMemoryConsumersGraphs();
	
    // initialize KO model
    function outputModel() {
        self = this;
        self.perfDiagObj = perfDiagObj;
	// Uncomment once the High Res Data flows into the output
		
        self.getDisksWithoutHighResData = function () {
            let allDisks = new Set(perfDiagObj.VolumeMapInfoCollector?.LatestData?.DiskMap?.map(d => d.BlockDevice))
            let highResDisks = new Set(perfDiagObj.HighResolutionDiskRule?.PerfDataPerDisk?.map(d => d.DiskName))
            let otherDisks = Array.from(allDisks).filter(d => !highResDisks?.has(d.slice(d.lastIndexOf('/') + 1)))
            return otherDisks.join(", ")
        }	
    }

    ko.bindingHandlers.numericText = {
        update: function (element, valueAccessor, allBindingsAccessor) {
            var value = ko.utils.unwrapObservable(valueAccessor()),
                precision = ko.utils.unwrapObservable(allBindingsAccessor().precision) || ko.bindingHandlers.numericText.defaultPrecision,
                unit = ko.utils.unwrapObservable(allBindingsAccessor().unit) || ko.bindingHandlers.numericText.defaultUnit,
                formattedValue = value.toFixed(precision) + ' ' + unit;

            ko.bindingHandlers.text.update(element, function () { return formattedValue; });
        },
        defaultPrecision: 1,
        defaultUnit: ''
    };

    ko.applyBindings(new outputModel());

    // setting up the filter textbox for the below tables
    var options = {
        tableIdsToFilter: [
            "eventsTable",
            "findingsTable",
            "netstatTable",
            "firewalldfirewallTable",
            "DriversTable",
	        "taskListTable",
	        "lvmmapTable",
	        "diskConfig_diskmapTable",
	        "ioStats5STable"
        ]
    };

    var optionsName = {
        tableNamesToFilter: [   
			"sqltable"
        ]
    };

    for (var i = 0; i < optionsName.tableNamesToFilter.length; i++) {
        var elements = document.getElementsByName(optionsName.tableNamesToFilter[i]);
        for (var j = 0; j < elements.length; j++) {
            setupFilter(elements[j].id);
        }
    }

    for (var i = 0; i < options.tableIdsToFilter.length; i++) {
        setupFilter(options.tableIdsToFilter[i]);
    }
    
    /*
    $(".accordion-toggle").on("click", function () {
        if ($(this).find(".glyphicon").hasClass("glyphicon-plus")) {
            $(this).find(".glyphicon").removeClass("glyphicon-plus").addClass("glyphicon-minus");
        } else {
            $(this).find(".glyphicon").removeClass("glyphicon-minus").addClass("glyphicon-plus");
        }
    });
    */
    
    $(".collapse").on("show.bs.collapse", function (e) {
        $(e.currentTarget.previousElementSibling).find(".glyphicon").removeClass("glyphicon-plus").addClass("glyphicon-minus");
    });
    $(".collapse").on("hide.bs.collapse", function (e) {
        $(e.currentTarget.previousElementSibling).find(".glyphicon").removeClass("glyphicon-minus").addClass("glyphicon-plus");
    });

    
    var highResCsvDataFileInput = document.getElementById("highResCsvDataFile");
    if (highResCsvDataFileInput != null) {
        highResCsvDataFileInput.addEventListener('change', function (e) {

            ShowCsvLoadErrorMsg(null);

            var file = highResCsvDataFileInput.files[0];
            var fileTable = document.getElementById("historicalHighResDiskDataTable");
            var foundFile = false;
            for (var i = 1; i < fileTable.rows.length; i++) {

                var cellContent = fileTable.rows[i].cells[0].innerHTML;
                if (cellContent.toLowerCase() == file.name.toLowerCase()) {
                    foundFile = true;
                }
            }
            if (!foundFile) {
                ShowCsvLoadErrorMsg(file.name + " is not the right CSV file for this report. Please load files listed in the table.");
                return;
            }

            var reader = new FileReader();
            reader.onload = function () {
                try {
                    var csvContent = reader.result;
                    var graph = new HighResolutionDiskUsageGraph();
                    graph.init(
                        perfDiagObj.HighResolutionDiskRule.AnalysisStartTime,
                        perfDiagObj.HighResolutionDiskRule.AnalysisEndTime,
                        perfDiagObj.HighResolutionDiskRule.Configuration.DetectWindowMS);
                    Csv2Json(csvContent, function (csvObj) { graph.onCsvLineRead(csvObj) });
                    graph.genGraph();

                    // after graphs are loaded, highlight the row with the right file name
                    var fileTable = document.getElementById("historicalHighResDiskDataTable");
                    var newScrollTop = 0;
                    var foundMatch = false;
                    // skip header
                    for (var i = 1; i < fileTable.rows.length; i++) {

                        var cellContent = fileTable.rows[i].cells[0].innerHTML;
                        if (cellContent.toLowerCase() == file.name.toLowerCase()) {
                            fileTable.rows[i].className = "highlight";
                            rowHighlighted = fileTable.rows[i];
                            foundMatch = true;
                        } else {
                            fileTable.rows[i].className = "";
                        }

                        if (!foundMatch) {
                            newScrollTop += fileTable.rows[i].offsetTop;
                        }
                    }
                    // scroll to the right row
                    var tbody = document.getElementById("historicalHighResDiskDataTableTBody");
                    tbody.scrollTop = newScrollTop;
                }
                catch (err) {
                    ShowCsvLoadErrorMsg(err.message);
                }
            };
            reader.readAsText(file);
        }, false);
    }    
}

function ShowCsvLoadErrorMsg(errMsg) {
    var errMsgDiv = document.getElementById("loadCsvErrMsg");
    if (errMsg) {
        errMsgDiv.innerHTML = errMsg;
        errMsgDiv.style.color = "Red";
    } else {
        errMsgDiv.innerHTML = "";
    }

}

function Csv2Json(csvContent, onCsvLineRead) {

    var lines = csvContent.split("\n");
    var headers = lines[0].split(",");

    for (var i = 1; i < lines.length; i++) {
        if (lines[i]) {
            var obj = {};
            var currentline = lines[i].split(",");
            if (currentline.length != headers.length) {
                throw new Error("Row " + i + " doesn't have the same number of columns as header.The content in row is: " + currentline);
            }
            for (var j = 0; j < headers.length; j++) {
                if (currentline[j] == null) {
                    throw new Error("Row " + i + " Column " + j + " doesn't have the valid value. The content in row is: " + currentline);
                }
                obj[headers[j].trim()] = currentline[j].trim();
            }
            onCsvLineRead(obj);
        }
    }
}

function OpenRuleResult(rule) {
    $(`#mainTabs a[href="#${rule.Category}Tab"]`).tab('show');
    $(`#${rule.Category}Tab a[href="#${rule.Category}Tab_Findings"]`).tab('show');
    return false;
}

function OpenRuleTab(ruleName) {
    var rules = perfDiagObj.RuleResults.filter(function (r) { return r.RuleName.toUpperCase() === ruleName.toUpperCase() });
    var targetRule = rules[0];
    return OpenRuleResult(targetRule);
}

function GetRuleTabTitle(ruleName) {
    var rules = perfDiagObj.RuleResults.filter(function (r) { return r.RuleName.toUpperCase() === ruleName.toUpperCase() });
    var targetRule = rules[0];
    return $(`#${targetRule.Category}Tab a[href="#${targetRule.Category}Tab_Findings"]`).text(); ``
}

function getLatencyClass(latency) {
    return latency < 10 ? 'info' : (latency < 50 ? 'warning' : 'danger');
}

function ShowFeedbackDialog(feedback) {
    var feedbackTitleText;
    if (feedback == "thumbsup") {
        feedbackTitleText = "Send us your feedback";
    } else if (feedback == "thumbsdown") {
        feedbackTitleText = "Report a problem";
    }
    $("#feedbacktitle").text(feedbackTitleText);

    $("#feedbackModal").on('hidden.bs.modal', function () {
        $("#feedbackdetails").val("");
    });

    $("#feedbackModal").modal();
}

function OpenFindingFeedback(finding, rule, feedback) {
    $("#feedbackPrompt").text('Please let us know about your experience with the finding of ' + rule.RuleName);

    $("#submitfeedbackbtn").click(function () {
        var fbDetails = $("#feedbackdetails").val();
        appInsights.trackEvent({
            name: "UserFeedbackOnRuleFinding",
            properties:
            {
                ReportId: perfDiagObj.ReportId,
                IsLinux: true,
                FindingId: finding.Id,
                RuleId: rule.RuleId,
                RuleName: rule.RuleName,
                Type: feedback,
                Details: fbDetails 
            }
        });

        $("#feedbackModal").modal('hide');
    });

    ShowFeedbackDialog(feedback);
}

function OpenOverallFeedback(feedback) {
    $("#feedbackPrompt").text('Please let us know about your experience with PerfInsights.');

    $("#submitfeedbackbtn").click(function () {
        var fbDetails = $("#feedbackdetails").val();

        appInsights.trackEvent({ name: "UserFeedbackOnReport", properties: { ReportId: perfDiagObj.ReportId, IsLinux: true, Type: feedback,Details: fbDetails } });

        $("#feedbackModal").modal('hide');
    });

    ShowFeedbackDialog(feedback);
}

function setupFilter(tableId) {

    var table = document.getElementById(tableId);

    // if the table is not found, return
    if (table == null)
        return;

    // if the table has less than 10 rows, then don't display the filter textbox
    var rowCount = $('#' + tableId + ' tr').length;
    if (rowCount < 10) {
        return;
    }

    // Constructing the filter textbox to be placed before the table
    // <input type="search" id="table1search" class="light-table-filter" data-id="table1"  placeholder="Filter" />
    var newInputElement = document.createElement("input");
    newInputElement.type = "search";
    newInputElement.id = tableId + "search";
    newInputElement.placeholder = "Filter";
    var arialabelattr = document.createAttribute("aria-label");
    arialabelattr.value = "Search " + tableId;
    newInputElement.setAttributeNode(arialabelattr);
    var att = document.createAttribute("data-id");
    att.value = tableId;
    newInputElement.setAttributeNode(att);

    $(table).before(newInputElement);

    var input = document.getElementById(newInputElement.id);
    $(input).after("<br />");
    if (tableId == "otherFindingsTable" || tableId == "blobFindingsTable") {
        $(input).after("<br />");
    }
    input.oninput = _onInputEvent;
}

function _onInputEvent(e) {
    var Arr = Array.prototype;
    _input = e.target;
    var table = document.getElementById(_input.getAttribute("data-id"));
    Arr.forEach.call(table.tBodies, function (tbody) {
        Arr.forEach.call(tbody.rows, _filter);
    });
}

function _filter(row) {
    // skip the filtering of the 1st header row
    if (row.innerHTML.indexOf("<th>") >= 0 && row.innerHTML.indexOf("<table") <= 0)
        return;
    var text = row.textContent.toLowerCase(), val = _input.value.toLowerCase();
    row.style.display = text.indexOf(val) === -1 ? 'none' : '';
}

function compareFindings(f1, f2) {
    const levelMap = { "High": 2, "Medium": 1, "Low": 0 }
    let impactLevelDiff = levelMap[f1.ImpactLevel] - levelMap[f2.ImpactLevel]
    let count = f => f?.ImpactedResources ?? 0
    return impactLevelDiff != 0 ? impactLevelDiff : count(f1) - count(f2)
}

function getFindingImpactLevelClass(impactLevel) {
    if (impactLevel == "High" ) {
        return "impact-critical";
    } else if (impactLevel == "Medium") {
        return "impact-warning";
    } else {
        return "impact-low";
    }
}

var has_loaded_perf_diag = false;
function loadPerfDiag() {
    if (has_loaded_perf_diag) {
        return;
    }

    perfDiagObj?.DataCollectors?.forEach(dataCollector => {
        let collectedData = dataCollector.CollectedData
        dataCollector.LatestData = collectedData.length > 0 ? collectedData[collectedData.length - 1].Data : null
        perfDiagObj[dataCollector.DataCollectorName] = dataCollector
    })
	
    let netstat_states = perfDiagObj?.NetworkMapInfoDC?.LatestData?.Netstat?.map(item => item.State) ?? []
	
    if (perfDiagObj?.NetworkMapInfoDC?.LatestData) {
        perfDiagObj.NetworkMapInfoDC.LatestData.NetstatSummary = { 
            'netstat_timewait_cnt' : netstat_states.filter(s => s === "TIME_WAIT").length,
            'netstat_listen_cnt' : netstat_states.filter(s => s === "LISTEN").length,
            'netstat_established_cnt' : netstat_states.filter(s => s === "ESTABLISHED").length,
            'netstat_stateless_cnt' : netstat_states.filter(s => s === "Stateless").length
        }
    }
    let compareRuleResults = (r1, r2) => {
        const tiers = [['High', 'Error'], ['Medium'], ['Low']]
        let impactCount = (r,tier) => r.RuleFindings?.filter(f => tier.includes(f.ImpactLevel)).length ?? 0
        for (tier of tiers) {
            let diff = impactCount(r1, tier) - impactCount(r2, tier)
            if (diff !== 0) {
                return diff
            }
        }
        return 0
    }

    let ruleResults = perfDiagObj?.RuleResults ?? []
    let categories = Array.from(new Set(ruleResults.map(rule => rule.Category).filter(c=>c)))
    categories.forEach(category => { perfDiagObj[category + "Rule"] = {
        RuleName: category + "Rule",
        RuleFindings: ruleResults
                        .filter(r => r.Category === category)
                        .map(r => r.RuleFindings)
                        .flat()
                        .sort(compareFindings)
                        .reverse()
    }})
    perfDiagObj.HighMemoryRule = perfDiagObj.RuleResults.find(obj => obj.RuleName === "HighMemoryRule")
    perfDiagObj.HighCpuRule = perfDiagObj.RuleResults.find(obj => obj.RuleName === "HighCpuRule")
    perfDiagObj.HighDiskRule = perfDiagObj.RuleResults.find(obj => obj.RuleName === "HighDiskRule")
    perfDiagObj.HighResolutionDiskRule = perfDiagObj.RuleResults.find(obj => obj.RuleName === "HighResolutionDiskRule")
    perfDiagObj.DiskConfigurationRule = perfDiagObj.RuleResults.find(obj => obj.RuleName === "DiskConfigurationRule")


    if (perfDiagObj ?? false) {
        perfDiagObj.TotalFindings = ruleResults.map(r => r.RuleFindings).flat().length
    }
    perfDiagObj?.RuleResults?.sort(compareRuleResults)?.reverse()
    has_loaded_perf_diag = true;
}



function rangeChanged(graph2d, options) {
    return function (properties) {
        if (properties.byUser != true) {
            return;
        }
        // only draw points when the current window in view is less than 10 seconds for performance
        if (properties.end.getTime() - properties.start.getTime() > 10000) {
            if (options.drawPoints.enabled == true) {
                options.drawPoints.enabled = false;
                options.start = properties.start;
                options.end = properties.end;
                graph2d.setOptions(options);
            }
        }
        else {
            if (options.drawPoints.enabled == false) {
                options.drawPoints.enabled = true;
                options.start = properties.start;
                options.end = properties.end;
                graph2d.setOptions(options);
            }
        }
    }
}

function customizeMinorLabels(date, scale, step) {
    var minorLabelsFormat = {
        millisecond: 'SSS',
        second: 's',
        minute: 'HH:mm',
        hour: 'HH:mm',
        weekday: 'ddd D',
        day: 'D',
        month: 'MMM',
        year: 'YYYY'
    };
    var minorLabelsUnit = {
        millisecond: 'ms',
        second: 's',
        minute: '',
        hour: '',
        weekday: '',
        day: '',
        month: '',
        year: ''
    };
    var format = minorLabelsFormat[scale];
    var unit = minorLabelsUnit[scale];
    return format && format.length > 0 ? vis.moment(date).format(format) + unit : '';
}


function HighResolutionDiskUsageGraph() {

    this.disksNumbers = {};
    this.dataForIOCountGraph = [];
    this.dataForIOSizeGraph = [];
    this.dataForIOLatencyGraph = [];
    this.disksGroupsForIOCount = new vis.DataSet();
    this.disksGroupsForIOSize = new vis.DataSet();
    this.disksGroupsForIOLatency = new vis.DataSet();
    this.maxIops = 0;
    this.maxMbps = 0;

    this.initFromJsonObj = function (analysisStartTime, analysisEndTime, detectWindowMS, PerfDataPerDisk) {

        this.init(analysisStartTime, analysisEndTime, detectWindowMS);

        for (var disk = 0; PerfDataPerDisk != null && disk < PerfDataPerDisk.length; disk++) {
            var diskNum = PerfDataPerDisk[disk].DiskNumber;
            var diskName = PerfDataPerDisk[disk].DiskName; 
            if (diskNum < 0) {
                continue;
            }
            this.disksNumbers[diskNum] = diskName; 
            for (var index = 0; index < PerfDataPerDisk[disk].DataPoints.length; index++) {
                var dp = PerfDataPerDisk[disk].DataPoints[index];
                dp.TM ??= 0
                dp.QL ??= 0
                dp.IC ??= 0
                dp.KB ??= 0
                dp.US50 ??= 0
                dp.US99 ??= 0

                // This formula converts IO and throughput from count during detect window to IO per second and MB per second respectively to display in graphs  
                var iops = dp.IC * (1000 / this.detectTimeWindowMS);
                var mbps = dp.KB * (1000 / this.detectTimeWindowMS) / 1000; // divide by 1000 converts megabytes (vs mebibytes)
                if (iops > this.maxIops) {
                    this.maxIops = iops;
                }
                if (mbps > this.maxMbps) {
                    this.maxMbps = mbps;
                }

                this.dataForIOCountGraph.push({ x: this.startTime + dp.TM, y: iops, group: diskNum * 2 });
                this.dataForIOCountGraph.push({ x: this.startTime + dp.TM, y: dp.QL, group: diskNum * 2 + 1 });

                this.dataForIOSizeGraph.push({ x: this.startTime + dp.TM, y: mbps, group: diskNum * 2 });
                this.dataForIOSizeGraph.push({ x: this.startTime + dp.TM, y: dp.QL, group: diskNum * 2 + 1 });

                this.dataForIOLatencyGraph.push({ x: this.startTime + dp.TM, y: dp.US50, group: diskNum * 3 });
                this.dataForIOLatencyGraph.push({ x: this.startTime + dp.TM, y: dp.US99, group: diskNum * 3 + 1 });
                this.dataForIOLatencyGraph.push({ x: this.startTime + dp.TM, y: dp.QL, group: diskNum * 3 + 2 });

                if (dp.TM < this.minRelativeTimeMS) {
                    this.minRelativeTimeMS = dp.TM;
                }
                if (dp.TM > this.maxRelativeTimeMS) {
                    this.maxRelativeTimeMS = dp.TM;
                }
            }
        }
    };

    this.init = function (analysisStartTime, analysisEndTime, detectWindowMS) {
        this.startTime = moment(analysisStartTime, "YYYY-MM-DD HH:mm:ss.SSS");
        this.endTime = moment(analysisEndTime, "YYYY-MM-DD HH:mm:ss.SSS");
        this.detectTimeWindowMS = detectWindowMS;
        this.minRelativeTimeMS = moment.duration(this.endTime.diff(this.startTime)).asMilliseconds();
        this.maxRelativeTimeMS = 0;
    };

    this.onCsvLineRead = function (dp) {
        var diskNum = +dp.DiskNumber;
        var diskName = dp.DiskName;
        if (diskNum < 0) {
            return;
        }
        var diskNumberExist = false;
        for (var i = 0; i < this.disksNumbers.length; i++) {
            if (this.disksNumbers[i] == diskNum) {
                diskNumberExist = true;
            }
        }
        if (!diskNumberExist) {
            this.disksNumbers[diskNum] = diskName;
        }

        var tm = +dp.TM;
        var ic = +dp.IC;
        var ql = +dp.QL;
        var kb = +dp.KB;
        var us50 = +dp.US50;
        var us99 = +dp.US99;

        // This formula converts IO and throughput from count during detect window to IO per second and MB per second respectively to display in graphs 
        var iops = dp.IC * (1000 / this.detectTimeWindowMS);
        var mbps = dp.KB * (1000 / this.detectTimeWindowMS) / 1000; // divide by 1000 converts megabytes (vs mebibytes)
        if (iops > this.maxIops) {
            this.maxIops = iops;
        }
        if (mbps > this.maxMbps) {
            this.maxMbps = mbps;
        }

        this.dataForIOCountGraph.push({ x: this.startTime + tm, y: iops, group: diskNum * 2 });
        this.dataForIOCountGraph.push({ x: this.startTime + tm, y: ql, group: diskNum * 2 + 1 });

        this.dataForIOSizeGraph.push({ x: this.startTime + tm, y: mbps, group: diskNum * 2 });
        this.dataForIOSizeGraph.push({ x: this.startTime + tm, y: ql, group: diskNum * 2 + 1 });

        
        // Uncomment when we're ready to show latency graphs 
        // this.dataForIOLatencyGraph.push({ x: this.startTime + tm, y: us50, group: diskNum * 3 });
        // this.dataForIOLatencyGraph.push({ x: this.startTime + tm, y: us99, group: diskNum * 3 + 1 });
        // this.dataForIOLatencyGraph.push({ x: this.startTime + tm, y: ql, group: diskNum * 3 + 2 });

        if (tm < this.minRelativeTimeMS) {
            this.minRelativeTimeMS = tm;
        }
        if (tm > this.maxRelativeTimeMS) {
            this.maxRelativeTimeMS = tm;
        }
    };

    this.genGraph = function () {
        for (const [key, value] of Object.entries(this.disksNumbers)) {
            var diskNum = key 
            var diskName = value

            if (diskNum < 0) {
                continue;
            }
            this.disksGroupsForIOCount.add({
                id: diskNum * 2,
                style: "stroke-width:2;stroke-dasharray:1 0;",
                content: '/dev/' + diskName + ' IOPS'
            });
            this.disksGroupsForIOCount.add({
                id: diskNum * 2 + 1,
                content: '/dev/' + diskName + ' QueueDepth',
                style: "stroke-width:2;stroke-dasharray:1 0;",
                options: {
                    yAxisOrientation: 'right'
                }
            });

            this.disksGroupsForIOSize.add({
                id: diskNum * 2,
                content: '/dev/' + diskName + ' Throughput(MBPS)',
                style: "stroke-width:2;stroke-dasharray:1 0;"
            });
            this.disksGroupsForIOSize.add({
                id: diskNum * 2 + 1,
                content: '/dev/' + diskName + ' QueueDepth',
                style: "stroke-width:2;stroke-dasharray:1 0;",
                options: {
                    yAxisOrientation: 'right'
                }
            });

            this.disksGroupsForIOLatency.add({
                id: diskNum * 3,
                content: '/dev/' + diskName + ' IOLatency-50th percentile(μs)',
                style: "stroke-width:2;stroke-dasharray:1 0;"
            });
            this.disksGroupsForIOLatency.add({
                id: diskNum * 3 + 1,
                content: '/dev/' + diskName + ' IOLatency-99th percentile(μs)',
                style: "stroke-width:2;stroke-dasharray:1 0;"
            });
            this.disksGroupsForIOLatency.add({
                id: diskNum * 3 + 2,
                content: '/dev/' + diskName + ' QueueDepth',
                style: "stroke-width:2;stroke-dasharray:1 0;",
                options: {
                    yAxisOrientation: 'right'
                }
            });
        }

        var maxIopsDataGroupId = this.disksGroupsForIOCount.max("id").id;
        var maxMbpsDataGroupId = this.disksGroupsForIOSize.max("id").id;
        var maxLatencyDataGroupId = this.disksGroupsForIOLatency.max("id").id;

        this.createFixedGroupsForThrottling();

        var datasetForIOCount = new vis.DataSet({
            type: { start: 'ISODate', end: 'ISODate' }
        });

        var datasetForIOSize = new vis.DataSet({
            type: { start: 'ISODate', end: 'ISODate' }
        });

        var datasetForIOLatency = new vis.DataSet({
            type: { start: 'ISODate', end: 'ISODate' }
        });

        datasetForIOCount.add(this.dataForIOCountGraph);
        datasetForIOSize.add(this.dataForIOSizeGraph);
        datasetForIOLatency.add(this.dataForIOLatencyGraph);

        var containerForIOCount = document.getElementById('highresdiskcount_visualization');
        containerForIOCount.innerHTML = '';
        var containerForIOSize = document.getElementById('highresdisksize_visualization');
        containerForIOSize.innerHTML = '';

        // Uncomment when we're ready to show latency graphs
        
        // var containerForIOLatency = document.getElementById('highresdisklatency_visualization');
        // containerForIOLatency.innerHTML = '';

        var optionsForIOCount = {
            sampling: true,
            drawPoints: { enabled: false, size: 3 },
            interpolation: false,
            legend: false,
            clickToUse: true,
            start: this.startTime,
            end: this.endTime,
            dataAxis: {
                left: { title: { text: "IOPS" } },
                right: { title: { text: "QueueDepth" } }
            },
            format: {
                minorLabels: customizeMinorLabels
            }
        };

        var optionsForIOSize = {
            sampling: true,
            drawPoints: { enabled: false, size: 3 },
            interpolation: false,
            legend: false,
            clickToUse: true,
            start: this.startTime,
            end: this.endTime,
            dataAxis: {
                left: { title: { text: "Throughput(MBPS)" } },
                right: { title: { text: "QueueDepth" } }
            },
            format: {
                minorLabels: customizeMinorLabels
            }
        };

        var optionsForIOLatency = {
            sampling: true,
            drawPoints: { enabled: false, size: 3 },
            interpolation: false,
            legend: false,
            clickToUse: true,
            start: this.startTime,
            end: this.endTime,
            dataAxis: {
                left: { title: { text: "IOLatency(μs)" } },
                right: { title: { text: "QueueDepth" } }
            },
            format: {
                minorLabels: customizeMinorLabels
            }
        };

        var ioCountGraph2d = new vis.Graph2d(containerForIOCount, datasetForIOCount, this.disksGroupsForIOCount, optionsForIOCount);
        ioCountGraph2d.setWindow(this.startTime + this.minRelativeTimeMS - 10000, this.startTime + this.maxRelativeTimeMS + 10000);
        populateExternalLegend(ioCountGraph2d, this.disksGroupsForIOCount, "highresdiskcount_legend", 0, maxIopsDataGroupId + 1, true);
        populateExternalLegend(ioCountGraph2d, this.disksGroupsForIOCount, "highresdiskcountthrottle_legend", maxIopsDataGroupId+1, this.disksGroupsForIOCount.max("id").id+1, true);
        ioCountGraph2d.on('rangechanged', rangeChanged(ioCountGraph2d, optionsForIOCount));
        addZoomButtons(containerForIOCount, function () { ioCountGraph2d.zoomIn(0.2); }, function () { ioCountGraph2d.zoomOut(0.2); });

        var ioSizeGraph2d = new vis.Graph2d(containerForIOSize, datasetForIOSize, this.disksGroupsForIOSize, optionsForIOSize);
        ioSizeGraph2d.setWindow(this.startTime + this.minRelativeTimeMS - 10000, this.startTime + this.maxRelativeTimeMS + 10000);
        populateExternalLegend(ioSizeGraph2d, this.disksGroupsForIOSize, "highresdisksize_legend", 0, maxMbpsDataGroupId + 1, true);
        populateExternalLegend(ioSizeGraph2d, this.disksGroupsForIOSize, "highresdisksizethrottle_legend", maxMbpsDataGroupId+1, this.disksGroupsForIOSize.max("id").id+1, true);
        ioSizeGraph2d.on('rangechanged', rangeChanged(ioSizeGraph2d, optionsForIOSize));
        addZoomButtons(containerForIOSize, function () { ioSizeGraph2d.zoomIn(0.2); }, function () { ioSizeGraph2d.zoomOut(0.2); });

        // Below is commented out until we are ready to show latency graphs

        // var ioLatencyGraph2d = new vis.Graph2d(containerForIOLatency, datasetForIOLatency, this.disksGroupsForIOLatency, optionsForIOLatency);
        // ioLatencyGraph2d.setWindow(this.startTime + this.minRelativeTimeMS - 10000, this.startTime + this.maxRelativeTimeMS + 10000);
        // populateExternalLegend(ioLatencyGraph2d, this.disksGroupsForIOLatency, "highresdisklatency_legend", 0, maxLatencyDataGroupId+1, true);
        // ioLatencyGraph2d.on('rangechanged', rangeChanged(ioLatencyGraph2d, optionsForIOLatency));
        // addZoomButtons(containerForIOLatency, function () { ioLatencyGraph2d.zoomIn(0.2); }, function () { ioLatencyGraph2d.zoomOut(0.2); });
    }

    this.createFixedGroupsForThrottling = function () {
        // the cap values are from https://azurewiki.cloudapp.net/StoragePerformanceCalculator 
        // and https://docs.microsoft.com/en-us/azure/storage/storage-scalability-targets#managed-virtual-machine-disks

        var azureDisksConfigs = [
            {
                Name: "P4",
                IOPSLimit: 120,
                ThroughputMBPS: 25
            },
            {
                Name: "P6",
                IOPSLimit: 240,
                ThroughputMBPS: 50
            },
            {
                Name: "Standard",
                IOPSLimit: 500,
                ThroughputMBPS: 60
            },
            {
                Name: "P10",
                IOPSLimit: 500,
                ThroughputMBPS: 100
            },
            {
                Name: "P15",
                IOPSLimit: 1100,
                ThroughputMBPS: 125
            },
            {
                Name: "P20",
                IOPSLimit: 2300,
                ThroughputMBPS: 150
            },
            {
                Name: "P30",
                IOPSLimit: 5000,
                ThroughputMBPS: 200
            },
            {
                Name: "P40/P50",
                IOPSLimit: 7500,
                ThroughputMBPS: 250
            },
            {
                Name: "P60",
                IOPSLimit: 12500,
                ThroughputMBPS: 480
            },
            {
                Name: "P70",
                IOPSLimit: 15000,
                ThroughputMBPS: 750
            },
            {
                Name: "P80",
                IOPSLimit: 20000,
                ThroughputMBPS: 750
            },
            {
                Name: "Ultra SSD",
                IOPSLimit: 160000,
                ThroughputMBPS: 2000
            }
        ];

        if (this.maxIops <= 0) {
            this.maxIops = azureDisksConfigs[0].IOPSLimit;
        }
        if (this.maxMbps <= 0) {
            this.maxMbps = azureDisksConfigs[0].ThroughputMBPS;
        }

        var iopsGroupIndexBase = this.disksGroupsForIOCount.max("id").id+1;
        var mbpsGroupIndexBase = this.disksGroupsForIOSize.max("id").id+1;
        for (var i = 0, j = 0; i < azureDisksConfigs.length; i++) {
            // Add throttling limit lines dynamically based on the max IOPS and MBPS values.
            if (azureDisksConfigs[i].IOPSLimit <= this.maxIops
                || azureDisksConfigs[i].IOPSLimit - this.maxIops <= azureDisksConfigs[i].IOPSLimit * 0.1
                || azureDisksConfigs[i].ThroughputMBPS <= this.maxMbps
                || azureDisksConfigs[i].ThroughputMBPS - this.maxMbps <= azureDisksConfigs[i].ThroughputMBPS * 0.1) {

                this.addThrottlingDataPoints(azureDisksConfigs[i], iopsGroupIndexBase + j, mbpsGroupIndexBase + j)
                j++;

            } else{
                // Add up to three additional throttling limit lines above the max iops/mbps
                for (var k = 0; i+k < 3; k++) {
                    this.addThrottlingDataPoints(azureDisksConfigs[i+k], iopsGroupIndexBase + j + k, mbpsGroupIndexBase + j + k)               
                }
                break;
            }
        }
    }

    this.addThrottlingDataPoints = function (azureDiskConfig, iopsGroupIndex, mbpsGroupIndex) {
        this.dataForIOCountGraph.push({ x: this.startTime, y: azureDiskConfig.IOPSLimit, group: iopsGroupIndex });
        this.dataForIOCountGraph.push({ x: this.endTime, y: azureDiskConfig.IOPSLimit, group: iopsGroupIndex });
        this.dataForIOSizeGraph.push({ x: this.startTime, y: azureDiskConfig.ThroughputMBPS, group: mbpsGroupIndex }); // MBPS
        this.dataForIOSizeGraph.push({ x: this.endTime, y: azureDiskConfig.ThroughputMBPS, group: mbpsGroupIndex }); // MBPS
        this.disksGroupsForIOCount.add({ id: iopsGroupIndex, style: "stroke-width:2;stroke-dasharray:10 10;", content: '<b>' + azureDiskConfig.Name + '</b> Disk IOPS Throttle Limit', visible: false });
        this.disksGroupsForIOSize.add({ id: mbpsGroupIndex, style: "stroke-width:2;stroke-dasharray:10 10;", content: '<b>' + azureDiskConfig.Name + '</b> Disk Throughput(MBPS) Throttle Limit', visible: false });
    }
}

function addZoomButtons(visualDiv, zoomInFunc, zoomOutFunc) {
    var zoombtn = document.createElement("div");
    zoombtn.setAttribute("class", "zoombuttons");
    var zoomIn = document.createElement("input");
    zoomIn.setAttribute("type", "button");
    zoomIn.setAttribute("value", "Zoom in");
    zoomIn.addEventListener("click", zoomInFunc);
    zoombtn.appendChild(zoomIn);
    var zoomOut = document.createElement("input");
    zoomOut.setAttribute("type", "button");
    zoomOut.setAttribute("value", "Zoom out");
    zoomOut.addEventListener("click", zoomOutFunc);
    zoombtn.appendChild(zoomOut);
    visualDiv.appendChild(zoombtn);
}

function genDefaultHighResolutionDiskUsageGraph() {
    if (!(perfDiagObj.HighResolutionDiskRule?.PerfDataPerDisk ?? false)) {
        return;
    }

    var graph = new HighResolutionDiskUsageGraph();
    graph.initFromJsonObj(
        perfDiagObj.HighResolutionDiskRule.AnalysisStartTime,
        perfDiagObj.HighResolutionDiskRule.AnalysisEndTime,
        perfDiagObj.HighResolutionDiskRule.Configuration.DetectWindowMS,
        perfDiagObj.HighResolutionDiskRule.PerfDataPerDisk);

    graph.genGraph();
}

/**
  * this function fills the external legend with content using the getLegend() function.
  */
function populateExternalLegend(graph2d, groups, legendName, startGroupIndex, endGroupIndex, interactive) {
    var legendDiv = document.getElementById(legendName);
    legendDiv.innerHTML = '';

    var checkbox = document.createElement('input');
    checkbox.type = 'checkbox';
    checkbox.id = legendName + "_checkbox";
    checkbox.checked = true;

    var label = document.createElement('label');
    label.htmlFor = legendName + "_checkbox";
    label.appendChild(document.createTextNode("Select All"));

    var br = document.createElement('br');

    legendDiv.appendChild(checkbox);
    legendDiv.appendChild(label);
    legendDiv.appendChild(br);

    // Keeps track of groups added to a legend and the number of groups that are hidden (unselected) at any given time. hiddenGroupsCount is to incremented or decremented as groups are made visible or hidden. A value of 0 indicated the Select All checkbox should be checked since no groups are hidden. 
    this.legendGroups = {
        groupsList: [],
        hiddenGroupsCount: 0
    }

    // get for all groups:
    groups.forEach(function(group) {
        if (group.id >= startGroupIndex && group.id < endGroupIndex) {
            // create divs
            var containerDiv = document.createElement("div");
            var iconDiv = document.createElement("div");
            var descriptionDiv = document.createElement("div");

            // give divs classes and Ids where necessary
            if (interactive) {
                containerDiv.className = 'legend-element-container';
            } else {
                containerDiv.className = 'legend-element-container-hidden';
            }
            containerDiv.id = group.id + "_legendContainer_" + legendName;
            iconDiv.className = "icon-container";
            descriptionDiv.className = "description-container";

            // get the legend for this group.
            var legend = graph2d.getLegend(group.id, 30, 30);

            // append class to icon. All styling classes from the vis.css/vis-timeline-graph2d.min.css have been copied over into the head here to be able to style the
            // icons with the same classes if they are using the default ones.
            legend.icon.setAttributeNS(null, "class", "legend-icon");

            // append the legend to the corresponding divs
            iconDiv.appendChild(legend.icon);
            descriptionDiv.innerHTML = legend.label;

            descriptionDiv.style.textAlign = "left";
            containerDiv.appendChild(iconDiv);
            containerDiv.appendChild(descriptionDiv);

            // append to the legend container div
            legendDiv.appendChild(containerDiv);

            legendGroups.groupsList.push(group); 
            if (interactive) {
                // bind click event to this legend element.
                containerDiv.onclick = toggleGraph.bind(this, group.id, graph2d, groups, legendName, legendGroups);
            }

            if (graph2d.isGroupVisible(group.id) == false) {
                legendGroups.hiddenGroupsCount++;
                containerDiv.className = containerDiv.className + "-hidden";
                checkbox.checked = false
            }
        }

    });
    checkbox.onclick = toggleSelectAll.bind(this, legendGroups, graph2d, groups, legendName);
}

function toggleSelectAll(groupsInLegend, graph2d, groups, legendName) {
    isChecked = document.querySelector('#' + legendName + "_checkbox").checked

    groupsInLegend.groupsList.forEach(function(group) { 
        var container = document.getElementById(group.id + "_legendContainer_" + legendName);
        if (isChecked && graph2d.isGroupVisible(group.id) == false) {
            // if checkbox is checked and group is hidden, show 
            showLegendGroup(group.id, groups, container)
        } else if (!isChecked && graph2d.isGroupVisible(group.id) == true)
            // if checkbox is not checked and group is visible, hide 
            hideLegendGroup(group.id, groups, container)
    })

    // If Select All is checked, reset the hiddenGroupsCount to 0. If unchecked, set to groupList.length since all groups are now hidden. 
    groupsInLegend.hiddenGroupsCount = isChecked ? 0 : groupsInLegend.groupsList.length;
}

/**
 * This function switchs the visible option of the selected group on an off.
 * @param groupId
 */
function toggleGraph(groupId, graph2d, groups, legendName, groupsInLegend) {
    // get the container that was clicked on.
    var container = document.getElementById(groupId + "_legendContainer_" + legendName);
    // if visible: hide and uncheck the legend checkbox
    if (graph2d.isGroupVisible(groupId) == true) {
        hideLegendGroup(groupId, groups, container)
        groupsInLegend.hiddenGroupsCount++;
    }
    else { // if invisible, show
        showLegendGroup(groupId, groups, container);
        groupsInLegend.hiddenGroupsCount--;
    }

    // If 0 groups are hidden, check Select All checkbox. Else uncheck it. 
    document.querySelector('#' + legendName + "_checkbox").checked = groupsInLegend.hiddenGroupsCount == 0 ? true : false;
}

function hideLegendGroup(groupId, groups, container) {
    groups.update({ id: groupId, visible: false });
        container.className = container.className + "-hidden";
}

function showLegendGroup(groupId, groups, container) {
    groups.update({ id: groupId, visible: true });
    container.className = container.className.replace("-hidden", "");
}

function HighCpuConsumersGraph() {

    this.spinners = [];
    this.spinPeriodsName = '';
    this.metricFieldName = '';
    this.elementId = '';

    this.initFromRuleResult = function (ruleResult, elementId, spinnerFieldName, spinPeriodsFieldName, metricFieldName) {
        this.spinners = ruleResult[spinnerFieldName];
        this.spinPeriodsFieldName = spinPeriodsFieldName;
        this.metricFieldName = metricFieldName;
        this.elementId = elementId;
        this.analysisStartTime = ruleResult.AnalysisStartTime;
        this.analysisEndTime = ruleResult.AnalysisEndTime;
    }

    this.genGraph = function () {

        var spinProcGroups = new vis.DataSet();
        var spinDataSet = new vis.DataSet({ type: { start: 'ISODate', end: 'ISODate' } });
        var data = [];
        var itemId = 0;
        for (var i = 0; i < this.spinners.length; i++) {
            var spinProc = this.spinners[i];

            var groupContent = '<b>' + spinProc.ProcessName + ' (' + spinProc.ProcessId + ')</b><ul>';
            if (spinProc.ServicesHosted) {
                groupContent += '<li>ServicesHosted: ' + spinProc.ServicesHosted + '</li>';
            }
            groupContent += '<li>Total CPU: ' + spinProc.TotalResourceUsageMS + 'ms</li>';
            groupContent += '<li>Avg CPU: ' + (+spinProc.TotalResourceUsagePercentage).toFixed(2) + '%</li>';
            groupContent += '</ul>';

            spinProcGroups.add({
                id: i,
                content: groupContent
            });

            var spinPeriods = spinProc[this.spinPeriodsFieldName];
            for (var j = 0; j < spinPeriods.length; j++) {
                var startTime = moment(spinPeriods[j].StartTime, "YYYY-MM-DD HH:mm:ss.SSS");
                var endTime = moment(spinPeriods[j].EndTime, "YYYY-MM-DD HH:mm:ss.SSS");
				var duration = moment.duration(endTime.diff(startTime));
				var cpuUsage = moment(spinPeriods[j].CpuUsageMS,);
				var avgCpu = moment(spinPeriods[j].AvgCpuPercentage);
                if (!isNaN(cpuUsage)) {
                    var content = 'Avg CPU: ' + avgCpu + '%; Total CPU: ' + cpuUsage + 'ms; Duration: ' + Math.floor(duration.asMilliseconds()) + 'ms';
                    var tooltip = '<p>Avg CPU: ' + avgCpu + '%</p><p>Total CPU: ' + cpuUsage + 'ms</p><p>Start time: ' + spinPeriods[j].StartTime + '</p><p>End time: ' + spinPeriods[j].EndTime + '</p><p>Duration: ' + Math.floor(duration.asMilliseconds()) + 'ms</p>';
                    data.push({
                        id: itemId++,
                        group: i,
                        start: startTime,
                        end: endTime,
                        content: content,
                        title: tooltip
                    });
                }
            }
        }
        spinDataSet.add(data);

        var container = document.getElementById(this.elementId);
        container.innerHTML = '';
        var minTimestamp = moment(this.analysisStartTime, "YYYY-MM-DD HH:mm:ss.SSS");
        minTimestamp.add(-1, "minutes");
        var maxTimestamp = moment(this.analysisEndTime, "YYYY-MM-DD HH:mm:ss.SSS");
        maxTimestamp.add(1, "minutes");
        var options = {
            clickToUse: true,
            min: minTimestamp,
            max: maxTimestamp
        };

        var timeline = new vis.Timeline(container, spinDataSet, spinProcGroups, options);

        addZoomButtons(container, function () { timeline.zoomIn(0.2); }, function () { timeline.zoomOut(0.2); });
    }
}

function genHighCpuConsumersGraphs() {
    if (perfDiagObj.HighCpuRule?.TopLongCpuSpinners?.length > 0) {

        var longRunningCpuGraph = new HighCpuConsumersGraph();
        longRunningCpuGraph.initFromRuleResult(
            perfDiagObj.HighCpuRule,
            'longrunningcpuconsumers_visualization',
            'TopLongCpuSpinners',
            'CpuSpinPeriods',
            'CpuUsageMS');
        longRunningCpuGraph.genGraph();
    }

    if (perfDiagObj.HighCpuRule?.TopSpikeCpuSpinners?.length > 0) {

        var spikeCpuGraph = new HighCpuConsumersGraph();
        spikeCpuGraph.initFromRuleResult(
            perfDiagObj.HighCpuRule,
            'spikecpuconsumers_visualization',
            'TopSpikeCpuSpinners',
            'CpuSpinPeriods',
            'CpuUsageMS');
        spikeCpuGraph.genGraph();
    }
}

function genHighProcessorsUsagePeriods() {
    if (!perfDiagObj?.HighCpuRule?.HighProcessorsUsagePeriods?.length) {

        return;
    }

    var highCpuRule = perfDiagObj.HighCpuRule;
    var firstPeriod = highCpuRule.HighProcessorsUsagePeriods[0];
    if (!firstPeriod?.ProcessorsUsage?.length <= 0) {
        return;
    }

    var processorGroups = new vis.DataSet();
    var processorUsageDataSet = new vis.DataSet({ type: { start: 'ISODate', end: 'ISODate' } });
    var data = [];
    var itemId = 0;
    for (var i = 0; i < firstPeriod.ProcessorsUsage.length; i++) {
        var cpuNumber = +(firstPeriod.ProcessorsUsage[i].CpuNumber);
        processorGroups.add({
            id: cpuNumber,
            content: '<b>Processor ' + cpuNumber + '</b>'
        });
    }
    for (var i = 0; i < highCpuRule.HighProcessorsUsagePeriods.length; i++) {
        var period = highCpuRule.HighProcessorsUsagePeriods[i];
        var startTime = moment(period.StartTime, "YYYY-MM-DD HH:mm:ss.SSS");
        var endTime = moment(period.EndTime, "YYYY-MM-DD HH:mm:ss.SSS");
        var duration = moment.duration(endTime.diff(startTime));
        for (var j = 0; j < period.ProcessorsUsage.length; j++) {
            var processorUsage = period.ProcessorsUsage[j];
            var cpuNum = +(processorUsage.CpuNumber);
            var usagePercent = parseFloat(processorUsage.ProcessorCpuPercentage).toFixed(2);
            var usageMS = +(processorUsage.ProcessorCpuTimeMS);
            var content = 'Total CPU: ' + usageMS + 'ms; Avg CPU: ' + usagePercent + '%';
            var tooltip = '<p>Total CPU: ' + usageMS + 'ms</p><p>Avg CPU: ' + usagePercent + '%</p><p>Start time: ' + period.StartTime + '</p><p>End time: ' + period.EndTime + '</p><p>Duration: ' + Math.floor(duration.asMilliseconds()) + 'ms</p>';
            data.push({
                id: itemId++,
                group: cpuNum,
                start: startTime,
                end: endTime,
                content: content,
                title: tooltip
            });
        }
    }

    processorUsageDataSet.add(data);
    var container = document.getElementById('highcpuperiods_visualization');
    container.innerHTML = '';
    var minTimestamp = moment(highCpuRule.AnalysisStartTime, "YYYY-MM-DD HH:mm:ss.SSS");
    minTimestamp.add(-1, "minutes");
    var maxTimestamp = moment(highCpuRule.AnalysisEndTime, "YYYY-MM-DD HH:mm:ss.SSS");
    maxTimestamp.add(1, "minutes");
    var options = {
        clickToUse: true,
        min: minTimestamp,
        max: maxTimestamp
    };

    var timeline = new vis.Timeline(container, processorUsageDataSet, processorGroups, options);
}

function genHighMemoryUsagePeriods() {
    if (!perfDiagObj?.HighMemoryRule?.HighMemoryUsagePeriods?.length) {
        return;
    }

    var highMemoryRule = perfDiagObj.HighMemoryRule;

    var memoryUsageDataSet = new vis.DataSet({ type: { start: 'ISODate', end: 'ISODate' } });
    var data = [];
    for (var i = 0; i < highMemoryRule.HighMemoryUsagePeriods.length; i++) {
        var period = highMemoryRule.HighMemoryUsagePeriods[i];
        var availablePercentage = parseFloat(period.AvailablePhysicalMemoryPercentage).toFixed(2);
        var startTime = moment(period.StartTime, "YYYY-MM-DD HH:mm:ss.SSS");
        var endTime = moment(period.EndTime, "YYYY-MM-DD HH:mm:ss.SSS");
        var content = 'Available Memory: ' + availablePercentage + '% ';
        var tooltip = '<p>Available Memory: ' + availablePercentage + '%</p>';
        data.push({
            id: i,
            start: startTime,
            end: endTime,
            content: content,
            title: tooltip
        });
    }

    memoryUsageDataSet.add(data);
    var container = document.getElementById('highmemoryperiods_visualization');
    container.innerHTML = '';
    var minTimestamp = moment(highMemoryRule.AnalysisStartTime, "YYYY-MM-DD HH:mm:ss.SSS");
    minTimestamp.add(-1, "minutes");
    var maxTimestamp = moment(highMemoryRule.AnalysisEndTime, "YYYY-MM-DD HH:mm:ss.SSS");
    maxTimestamp.add(1, "minutes");
    var options = {
        clickToUse: true,
        min: minTimestamp,
        max: maxTimestamp,
        minHeight: 200
    };

    var timeline = new vis.Timeline(container, memoryUsageDataSet, options);
}

function HighMemoryConsumersGraph() {

    this.spinners = [];
    this.spinPeriodsName = '';
    this.metricFieldName = '';
    this.elementId = '';

    this.initFromRuleResult = function (ruleResult, elementId, spinnerFieldName) {
        this.spinners = ruleResult[spinnerFieldName];
        this.elementId = elementId;
        this.analysisStartTime = ruleResult.AnalysisStartTime;
        this.analysisEndTime = ruleResult.AnalysisEndTime;
    }

    this.genGraph = function () {

        var spinProcGroups = new vis.DataSet();
        var spinDataSet = new vis.DataSet({ type: { start: 'ISODate', end: 'ISODate' } });
        var data = [];
        var itemId = 0;
        for (var i = 0; i < this.spinners.length; i++) {
            var spinProc = this.spinners[i];

            var groupContent = '<b>' + spinProc.ProcessName + ' (' + spinProc.ProcessId + ')</b><ul>';
            if (spinProc.ServicesHosted) {
                groupContent += '<li>ServicesHosted: ' + spinProc.ServicesHosted + '</li>';
            }
            groupContent += '<li>Avg Resident Memory: ' + formatSizeToHumanReadable(spinProc.AvgResidentMemoryKB, 'kb') + '</li>';
            groupContent += '<li>Avg Swap Memory: ' + formatSizeToHumanReadable(spinProc.AvgSwapMemoryKB, 'kb') + '</li>';
            groupContent += '</ul>';

            spinProcGroups.add({
                id: i,
                content: groupContent
            });

            var spinPeriods = spinProc.MemorySpinPeriods;
            for (var j = 0; j < spinPeriods.length; j++) {
                var startTime = moment(spinPeriods[j].StartTime, "YYYY-MM-DD HH:mm:ss.SSS");
                var endTime = moment(spinPeriods[j].EndTime, "YYYY-MM-DD HH:mm:ss.SSS");
                var period = spinPeriods[j];
                var duration = moment.duration(endTime.diff(startTime));
                var content = 'Avg Resident Memory: ' + formatSizeToHumanReadable(period.AvgResidentMemoryKB, 'kb')  + '; Avg Swap Memory: ' + formatSizeToHumanReadable(period.AvgSwapMemoryKB, 'kb');
                var tooltip = '<p>Avg Resident Memory (KB): ' + period.AvgResidentMemoryKB + '</p><p>Avg Swap Memory (KB): ' + period.AvgSwapMemoryKB + '</p><p>Start time: ' + spinPeriods[j].StartTime + '</p><p>End time: ' + spinPeriods[j].EndTime + '</p><p>Duration: ' + Math.floor(duration.asMilliseconds()) + 'ms</p>';
                data.push({
                    id: itemId++,
                    group: i,
                    start: startTime,
                    end: endTime,
                    content: content,
                    title: tooltip
                });
            }
        }
        spinDataSet.add(data);

        var container = document.getElementById(this.elementId);
        container.innerHTML = '';
        var minTimestamp = moment(this.analysisStartTime, "YYYY-MM-DD HH:mm:ss.SSS");
        minTimestamp.add(-1, "minutes");
        var maxTimestamp = moment(this.analysisEndTime, "YYYY-MM-DD HH:mm:ss.SSS");
        maxTimestamp.add(1, "minutes");
        var options = {
            clickToUse: true,
            min: minTimestamp,
            max: maxTimestamp
        };

        var timeline = new vis.Timeline(container, spinDataSet, spinProcGroups, options);

        addZoomButtons(container, function () { timeline.zoomIn(0.2); }, function () { timeline.zoomOut(0.2); });
    }
}

function genHighMemoryConsumersGraphs() {
    if (perfDiagObj.HighMemoryRule?.TopMemorySpinners?.length > 0) {

        var longRunningMemoryGraph = new HighMemoryConsumersGraph();
        longRunningMemoryGraph.initFromRuleResult(
            perfDiagObj.HighMemoryRule,
            'longrunningmemoryconsumers_visualization',
            'TopMemorySpinners');
        longRunningMemoryGraph.genGraph();
    }
}

function getApplicationInsightsEndpoint(){
    perfDiagObj.InstanceMetadataInfoDC ?? loadPerfDiag()
    switch (perfDiagObj.InstanceMetadataInfoDC?.LatestData?.Compute?.azEnvironment) {
        case 'AzureUSGovernmentCloud': return 'https://dc.applicationinsights.us/v2/track'
        case 'USNat': return 'https://dc.applicationinsights.azure.eaglex.ic.gov/v2/track'
        case 'USSec': return 'https://dc.applicationinsights.azure.microsoft.scloud/v2/track'
        default: return 'https://dc.services.visualstudio.com/v2/track'
    } 
}

function getApplicationInsightsKey(){
    perfDiagObj.InstanceMetadataInfoDC ?? loadPerfDiag()
    switch (perfDiagObj.InstanceMetadataInfoDC?.LatestData?.Compute?.azEnvironment) {
        case 'AzureUSGovernmentCloud': return '467f482c-20f9-fa17-99c2-5251d19f6114'
        case 'USNat': return 'a24fac76-a4cc-4435-9de4-0a212da7a8bd'
        case 'USSec': return 'ea44baa5-8065-47be-91a6-f82415f7046b'
        default: return '9561b95d-d813-42d5-98a0-3177b3290094'
    }
}

function isArm64() {
    vmarch = perfDiagObj.MachineConfigurationCollector?.LatestData?.VMArchitecture
    return vmarch == 'aarch64' || vmarch == 'arm64'
}

function formatSizeToHumanReadable(size, unit) {
    const humanReadableUnits = ['bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'];

    // Convert input to bytes
    let sizeInBytes = 0;
    try {
      if (unit.toLowerCase() === 'kb') {
          sizeInBytes = parseFloat(size) * 1024;
      } else if (unit.toLowerCase() === 'mb') {
          sizeInBytes = parseFloat(size) * 1024 * 1024;
      } else if (unit.toLowerCase() === 'gb') {
          sizeInBytes = parseFloat(size) * 1024 * 1024 * 1024;
      } else if (unit.toLowerCase() === 'bytes') {
          sizeInBytes = parseFloat(size);
      } else {
          throw new Error("formatSizeToHumanReadable(): Unit of '" + unit + "' is not recognized.");  
      }
    } catch(e) {
        console.log(e.message);
        return size + ' ' + unit;
    }
    
    // Format size as human-readable string
    if (sizeInBytes === 0) {
        return '0 bytes';
    } else {
        const digitGroups = Math.floor(Math.log10(sizeInBytes) / Math.log10(1024));
        if (digitGroups >= humanReadableUnits.findIndex((u) => u.toLowerCase() === unit.toLowerCase())) {
          const value = parseFloat((sizeInBytes / Math.pow(1024, digitGroups)).toFixed(2));
          const unit = humanReadableUnits[digitGroups];
          return value + ' ' + unit;
        } else {
          return size + ' ' + unit.toUpperCase()
        }
    }
}

  
// hook in the window onload handler as init()
window.onload = init;
